# Copyright (C) 2022 Tony Fischer
#
# This file is part of Bio Matrix Topology (BiMaTo).
#
# Bio Matrix Topology (BiMaTo) is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Bio Matrix Topology (BiMaTo) is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Bio Matrix Topology (BiMaTo).  If not, see <http://www.gnu.org/licenses/>.

'''The module :mod:`bimato.core` contains some fundamental functions for the `bimato` project. Mainly it contains the
custom binarization technique tailored towards 3D biompolymer network images and others.

See documentation for a general user guide.
'''


__author__ = "Tony Fischer (tku137)"
__copyright__ = "Copyright 2022, Tony Fischer (tku137)"
__license__ = "GPLv3"
__email__ = "tonyfischer@mailbox.org"
__status__ = "Development"
__version__ = "2022.1"
__credits__ = ["Tony Fischer (tku137)", "Alexander Hayn"]


from skimage.filters import threshold_local  # pylint: disable=no-name-in-module
from skimage.morphology import binary_closing, disk, ball, remove_small_objects
from scipy.ndimage.morphology import distance_transform_edt as edt

import numpy as np


def get_binary(data, bw_closing_frac=1024, min_obj_frac=256, win_frac=8, thresh_shift_percentage=0.01, morph_3d=True):
    '''Function to get a segmentation of input biopolymer network image data. Parameters are optional, but should be
    considered. Optimal values can vary for drastically differing image data.

    Notes
    -----
    All Parameters are determined automatically depending on input image properties. However, they can be influenced by
    specifying a factor.

    Parameters
    ----------
    data : numpy.ndarray
        Input image data
    bw_closing_frac : int
        Factor for calculation of the ball size during binary closing morphological operations
    min_obj_frac : int
        Factor for determining the minimum object size
    win_frac : int
        Factor for determining the local adaptive threshold window size
    thresh_shift_percentage : float
        Percentage factor in [0,1] specifying the amount the threshold should be shifted
    morph_3d : bool
        If morphological operations should be applied in 2D or 3D, default is 3D

    Returns
    -------
    numpy.ndarray
        Binary segmentation output

    '''

    # calc the threshold shift
    thresh_shift = thresh_shift_percentage * data.max()

    # determine optimal parameters
    small_edge_size = np.min(data.shape[:2])  # size of the smaller dim of an x-y image
    threshold_win = 2 * int(small_edge_size / (win_frac * 2)) + 1  # threshold_local window
    bw_closing_size = int(np.ceil(small_edge_size / bw_closing_frac))  # ball size for morphological closing

    min_obj_exp = 3 if morph_3d else 2  # wanna morph in 3D or 2D?
    min_obj = int(np.ceil(small_edge_size / min_obj_frac) ** min_obj_exp)  # objects must be this large

    # per plane adaptive threshold
    binary = np.zeros_like(data, dtype=np.bool)
    for z in range(data.shape[2]):
        # get the local threshold map
        t = threshold_local(data[..., z], threshold_win)
        # threshold the image with it
        binary[..., z] = data[..., z] > (t + thresh_shift)
        if not morph_3d:
            # do morphological operation in 2D, because 3D would cut of the bw_closing_size size in z direction!
            binary[..., z] = binary_closing(binary[..., z], footprint=disk(bw_closing_size))

    # morphological closing in 3D
    if morph_3d:
        binary = binary_closing(binary, footprint=ball(bw_closing_size))

    # remove small objects and holes
    binary = remove_small_objects(binary, min_size=min_obj)

    #
    return binary


def get_edm(binary, sampling):
    '''It takes a binary image and a dictionary of sampling rates and returns the Euclidean distance map

    Parameters
    ----------
    binary : numpy.ndarray
        a 3D binary array
    sampling : dict
        a dictionary with keys 'x', 'y', and 'z' that contain the voxel size in each dimension

    Returns
    -------
    numpy.ndarray
        The Euclidean distance map (EDM) of the binary image.

    '''
    edm = edt(
        np.logical_not(binary),
        sampling=[sampling['x'], sampling['y'], sampling['z']]
    )
    return edm



# Copyright (C) 2022 Tony Fischer
#
# This file is part of Bio Matrix Topology (BiMaTo).
#
# Bio Matrix Topology (BiMaTo) is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Bio Matrix Topology (BiMaTo) is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Bio Matrix Topology (BiMaTo).  If not, see <http://www.gnu.org/licenses/>.

'''The module :mod:`bimato` is the top-level namespace. It contains the following sub-modules:

- :mod:`bimato.core`: fundamental functions to the bimato project, mainly the custom binarization
- :mod:`bimato.poresize`: pore-size algorithm published at https://www.nature.com/articles/s41598-019-44764-5
- :mod:`bimato.network`: algorithms to analyze the network structure
- :mod:`bimato.utils`: various statistical and other utility functions

See documentation for a general user guide.
'''


__author__ = "Tony Fischer (tku137)"
__copyright__ = "Copyright 2022, Tony Fischer (tku137)"
__license__ = "GPLv3"
__email__ = "tonyfischer@mailbox.org"
__status__ = "Development"
__version__ = "2022.1"
__credits__ = ["Tony Fischer (tku137)", "Alexander Hayn"]


from bimato.api import *  # noqa: F401,F403



# Copyright (C) 2022 Tony Fischer
#
# This file is part of Bio Matrix Topology (BiMaTo).
#
# Bio Matrix Topology (BiMaTo) is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Bio Matrix Topology (BiMaTo) is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Bio Matrix Topology (BiMaTo).  If not, see <http://www.gnu.org/licenses/>.

"""
docstring of module
"""


__author__ = "Tony Fischer (tku137)"
__copyright__ = "Copyright 2022, Tony Fischer (tku137)"
__license__ = "GPLv3"
__email__ = "tonyfischer@mailbox.org"
__status__ = "Development"
__version__ = "2022.1"
__credits__ = ["Tony Fischer (tku137)", "Alexander Hayn"]

__all__ = ['core', 'poresize', 'network', 'utils',
           'get_binary', 'get_fiber_thickness', 'get_pore_sizes']


from . import core
from . import poresize
from . import network
from . import utils

from .core import get_binary
from .poresize import get_fiber_thickness, get_pore_sizes



# Copyright (C) 2022 Tony Fischer
#
# This file is part of Bio Matrix Topology (BiMaTo).
#
# Bio Matrix Topology (BiMaTo) is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Bio Matrix Topology (BiMaTo) is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Bio Matrix Topology (BiMaTo).  If not, see <http://www.gnu.org/licenses/>.

'''The module :mod:`bimato.poresize` contains the complete pore-size algorithm published at
https://www.nature.com/articles/s41598-019-44764-5. The most important is :func:`bimato.poresize.get_pore_sizes` and
:func:`bimato.poresize.get_fiber_thickness`.

See documentation for a general user guide.
'''


__author__ = "Tony Fischer (tku137)"
__copyright__ = "Copyright 2022, Tony Fischer (tku137)"
__license__ = "GPLv3"
__email__ = "tonyfischer@mailbox.org"
__status__ = "Development"
__version__ = "2022.1"
__credits__ = ["Tony Fischer (tku137)", "Alexander Hayn"]


from itertools import product
from skimage.feature import peak_local_max
from skimage.filters import gaussian  # pylint: disable=no-name-in-module
from skimage.morphology import ball
import numpy as np
import pandas as pd
from .core import get_edm
from .utils import get_voxel_volume, get_intervals


def calc_free_pore_space(df):
    '''Calculates the fraction of pore volume to sample volume.

    Parameters
    ----------
    df : pandas.DataFrame
        Input DataFrame that came from initial pore detection function

    Returns
    -------
    pandas.Series
        Pandas Series with calculated free pore space

    '''
    sum_pore_vol = df.loc[:, "Pore Volume [µm³]"].sum()
    cube_vol = df.loc[:, "Cube Volume [µm³]"].iloc[0]
    return sum_pore_vol / cube_vol


def get_pore_sizes(binary, sampling, sigma_frac=128, residual_pore_detection=False):
    '''This function takes a binary image of a 3D biopolymer network image and calculates the pore sizes given the
    pixel-micron conversion.

    Parameters
    ----------
    binary
        The binary image containing the segmentation of fluid (=0) and polymer phase (=1)
    sampling
        The physical size of each voxel in the image
    sigma_frac, optional
        The sigma parameter for the gaussian filter. The smaller the value, the more pores are detected
    residual_pore_detection, optional
        If True, the algorithm will try to detect even smaller and obscure pores

    Returns
    -------
        A dataframe with the following columns, each row represents data for a single pore:
        x [px] : x coordinate in pixel
        y [px] : y coordinate in pixel
        z [px] : z coordinate in pixel
        Diameter [µm] : pore diameter
        Residual Degree : number of iterative pore detections, 0 for first step, 1 for second
        Size x [px] : total x size of 3D image
        Size y [px] : total y size of 3D image
        Size z [px] : total z size of 3D image
        PhysicalSize x : pixel-micron conversion factor used
        PhysicalSize y : pixel-micron conversion factor used
        PhysicalSize z : pixel-micron conversion factor used
        Number Of Pores : total number of detected pores in sample
        Cube Volume [µm³] : physical volume of 3D image
        Single Pore Volume [µm³] : volume of detected pore
        Real Pore Volume [µm³] : physical volume occupied by detected pores, without overlapping of individual pores
        Collagen Volume [µm³] : volume of the polymer phase
        Fluid Volume [µm³] : volume of the fluid phase
        Single Pore Volume Fraction : space occupied in relation to the sample volume
        Real Pore Volume Fraction : space occupied in relation to the sample volume
        Collagen Volume Fraction : space occupied in relation to the sample volume
        Fluid Volume Fraction : space occupied in relation to the sample volume
        Zeta Single Pores : pore diameter scaled with its corresponding column fraction
        Zeta Real Pores : pore diameter scaled with its corresponding column fraction
        Zeta Collagen : pore diameter scaled with its corresponding column fraction
        Zeta Fluid : pore diameter scaled with its corresponding column fraction
        Pseudo Pore Diameter [µm] : theoretical pore diameter calculated by assuming that all detected pores would occupy the fluid-phase fully

    '''
    # calc fluid phase volume
    voxel_volume = get_voxel_volume(sampling)
    voxel_radius = np.cbrt((3*voxel_volume)/(4*np.pi))

    # calculate the EDM
    small_edge_size = np.min(binary.shape[:2])  # size of the smaller dim of an x-y image
    sigma = small_edge_size / sigma_frac
    edm = get_edm(binary, sampling)  # edm has microns as values

    # get pore coordinates
    pore_coords = peak_local_max(gaussian(edm, sigma=sigma))

    # construct DataFrame with diameter and other infos
    df_pores = pd.DataFrame(pore_coords, columns=['x [px]', 'y [px]', 'z [px]'])  # coords are px coords!
    # we take values from edm, so the result is microns again!
    df_pores['Diameter [µm]'] = edm[pore_coords[:, 0], pore_coords[:, 1], pore_coords[:, 2]] * 2  # Radius! So x2!
    # TODO: multi-degree
    df_pores['Residual Degree'] = 0
    # calc real pore volume
    pore_stack = np.zeros_like(binary)

    for _, pore in df_pores.iterrows():
        x, y, z, d, _ = pore.astype(np.int)  # need integers to index the array
        r = int((d / 2) / float(voxel_radius))  # convert to pixel-radius

        # construct coordinate arrays of a sphere at (x,y,z) with radius r
        pore_indices = np.argwhere(ball(r))
        pore_indices[:, 0] += x - r
        pore_indices[:, 1] += y - r
        pore_indices[:, 2] += z - r

        # filter out of range pore voxel
        pore_indices = pore_indices[pore_indices[:, 0] < pore_stack.shape[0]]
        pore_indices = pore_indices[pore_indices[:, 1] < pore_stack.shape[1]]
        pore_indices = pore_indices[pore_indices[:, 2] < pore_stack.shape[2]]

        # draw sphere
        pore_stack[pore_indices[:, 0], pore_indices[:, 1], pore_indices[:, 2]] = 1

    # residual poresize detection
    if residual_pore_detection:

        # this is collagen plus already detected pores
        residual_fluid = binary + pore_stack

        # calculate the EDM
        edm_residual = get_edm(residual_fluid, sampling)  # edm has microns as values

        # get pore coordinates
        pore_coords_residual = peak_local_max(gaussian(edm_residual, sigma=sigma))

        # construct DataFrame with diameter and other infos
        df_pores_residual = pd.DataFrame(pore_coords_residual, columns=['x [px]', 'y [px]', 'z [px]'])  # coords are px!
        # we take values from edm, so the result is microns again!
        df_pores_residual['Diameter [µm]'] = edm_residual[
                                                 pore_coords_residual[:, 0],
                                                 pore_coords_residual[:, 1],
                                                 pore_coords_residual[:, 2]] * 2  # Radius! So x2!
        df_pores_residual['Residual Degree'] = 1

        # calc real pore volume
        for _, pore in df_pores_residual.iterrows():  # TODO: refactor to function
            x, y, z, d, _ = pore.astype(np.int)  # need integers to index the array
            r = int((d / 2) / float(voxel_radius))  # convert to pixel-radius
            pore_indices = np.argwhere(ball(r))
            pore_indices[:, 0] += x - r
            pore_indices[:, 1] += y - r
            pore_indices[:, 2] += z - r
            pore_indices = pore_indices[pore_indices[:, 0] < pore_stack.shape[0]]
            pore_indices = pore_indices[pore_indices[:, 1] < pore_stack.shape[1]]
            pore_indices = pore_indices[pore_indices[:, 2] < pore_stack.shape[2]]
            pore_stack[pore_indices[:, 0], pore_indices[:, 1], pore_indices[:, 2]] = 1

        df_pores = pd.concat([df_pores, df_pores_residual], axis=0)

    # parsed meta-data
    # TODO: meta-data in comment lines above DataFrame
    size_x, size_y, size_z = binary.shape
    df_pores['Size x [px]'] = size_x
    df_pores['Size y [px]'] = size_y
    df_pores['Size z [px]'] = size_z
    df_pores['PhysicalSize x'] = sampling['x']
    df_pores['PhysicalSize y'] = sampling['y']
    df_pores['PhysicalSize z'] = sampling['z']

    # number of pores
    df_pores["Number Of Pores"] = df_pores.shape[0]

    # sample volume (the image cube)
    df_pores["Cube Volume [µm³]"] = (df_pores['Size x [px]'] * df_pores['Size y [px]'] * df_pores[
        'Size z [px]']) * voxel_volume

    # volume of each individual pore, the volumes might overlap!
    df_pores['Single Pore Volume [µm³]'] = (4 / 3) * np.pi * ((df_pores["Diameter [µm]"] / 2) ** 3)

    # calculate the real volume that all pores occupy
    df_pores['Real Pore Volume [µm³]'] = pore_stack.sum() * voxel_volume

    # calculate volume of collagen and fluid phase
    df_pores['Collagen Volume [µm³]'] = binary.sum() * voxel_volume
    df_pores['Fluid Volume [µm³]'] = df_pores["Cube Volume [µm³]"] - df_pores['Collagen Volume [µm³]']

    # how much space each component volume occupies in relation to the sample (cube) volume
    df_pores["Single Pore Volume Fraction"] = df_pores["Single Pore Volume [µm³]"] / df_pores["Cube Volume [µm³]"]
    df_pores["Real Pore Volume Fraction"] = df_pores["Real Pore Volume [µm³]"] / df_pores["Cube Volume [µm³]"]
    df_pores["Collagen Volume Fraction"] = df_pores["Collagen Volume [µm³]"] / df_pores["Cube Volume [µm³]"]
    df_pores["Fluid Volume Fraction"] = df_pores["Fluid Volume [µm³]"] / df_pores["Cube Volume [µm³]"]

    # the actual scaled pore-sizes Zeta. this is the pore diameter scaled with its corresponding column fraction.
    df_pores["Zeta Single Pores"] = df_pores["Diameter [µm]"] * df_pores["Single Pore Volume Fraction"]
    df_pores["Zeta Real Pores"] = df_pores["Diameter [µm]"] * df_pores["Real Pore Volume Fraction"]
    df_pores["Zeta Collagen"] = df_pores["Diameter [µm]"] * df_pores["Collagen Volume Fraction"]
    df_pores["Zeta Fluid"] = df_pores["Diameter [µm]"] * df_pores["Fluid Volume Fraction"]

    # pseudo pore diameter
    df_pores["Pseudo Pore Diameter [µm]"] = np.cbrt(
        (6 * (df_pores["Fluid Volume [µm³]"] / df_pores["Number Of Pores"])) / np.pi)

    #
    return df_pores


def get_fiber_thickness(binary, sampling, sigma_frac=512):
    '''This function takes a binary image of a 3D biopolymer network image and returns the fiber thickness

    Parameters
    ----------
    binary
        The binary image containing the segmentation of fluid (=0) and polymer phase (=1)
    sampling
        The physical size of each voxel in the image
    sigma_frac, optional
        The sigma parameter for the gaussian filter. The smaller the value, the more fibers are detected

    '''

    # calculate the EDM
    small_edge_size = np.min(binary.shape[:2])  # size of the smaller dim of an x-y image
    sigma = small_edge_size / sigma_frac
    edm = get_edm(np.logical_not(binary), sampling)  # edm has microns as values

    # get fiber coordinates
    fiber_coords = peak_local_max(gaussian(edm, sigma=sigma))

    # construct DataFrame with diameter and other infos
    df_fibers = pd.DataFrame(fiber_coords, columns=['x [px]', 'y [px]', 'z [px]'])  # coords are px!
    # we take values from edm, so the result is microns again!
    df_fibers['Diameter [µm]'] = edm[fiber_coords[:, 0], fiber_coords[:, 1], fiber_coords[:, 2]] * 2  # Radius! So x2!

    return df_fibers


def get_fragmented_poresize(binary, sampling, part_size_micron, residual_pore_detection=False, sigma_frac=128):
    data_shape = binary.shape
    intervals_x, intervals_y, intervals_z = get_intervals(data_shape, sampling, part_size_micron)

    df = list()

    for ix, iy, iz in product(intervals_x, intervals_y, intervals_z):
        # each interval (ix...) contains start and stop indices of the respective slice interval
        # cube[ interval[start] : interval[end], ... ) is the correct thing
        cube_slice = binary[ix[0]:ix[1], iy[0]:iy[1], iz[0]:iz[1]]

        # construct DataFrame with measures
        tmp_df = get_pore_sizes(cube_slice, sampling, sigma_frac, residual_pore_detection)

        tmp_df['ix_start'] = ix[0]
        tmp_df['ix_end'] = ix[1]
        tmp_df['iy_start'] = iy[0]
        tmp_df['iy_end'] = iy[1]
        tmp_df['iz_start'] = iz[0]
        tmp_df['iz_end'] = iz[1]
        tmp_df['Slice ID'] = f"{ix[0]}-{ix[1]}_{iy[0]}-{iy[1]}_{iz[0]}-{iz[1]}"

        # save data
        df.append(tmp_df)

    df = pd.concat(df, axis=0, ignore_index=True)

    return df



# Copyright (C) 2022 Tony Fischer
#
# This file is part of Bio Matrix Topology (BiMaTo).
#
# Bio Matrix Topology (BiMaTo) is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Bio Matrix Topology (BiMaTo) is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Bio Matrix Topology (BiMaTo).  If not, see <http://www.gnu.org/licenses/>.

"""
# TODO: write helpers module docstring
docstring of module
"""


__author__ = "Tony Fischer (tku137)"
__copyright__ = "Copyright 2022, Tony Fischer (tku137)"
__license__ = "GPLv3"
__email__ = "tonyfischer@mailbox.org"
__status__ = "Development"
__version__ = "2022.1"
__credits__ = ["Tony Fischer (tku137)", "Alexander Hayn"]


import numpy as np
from skimage.restoration import denoise_tv_chambolle


def read_lif_image(lif_image):
    '''It reads the image data from the readlif LifFile and returns it as a 3D numpy array

    Parameters
    ----------
    lif_image : readlif.reader.LifFile
        instance of readlif LifFile

    Returns
    -------
    numpy.ndarray
        A 3D numpy array of the image data.

    '''

    dims = lif_image.info['dims']

    data = np.zeros((dims.y, dims.x, dims.z))
    for z, im_plane in enumerate(lif_image.get_iter_z(t=0, c=0)):
        data[..., z] = np.array(im_plane)

    return data


def normalize_image(image, scale):
    """
    Standard function to normalize image data values to [0,1].

    Parameters
    ----------
    image : numpy.ndarray
        Input image data
    scale : int or float
        Scaling factor, e.g. 255 would scale to 8bit integer matrices

    Returns
    -------
    numpy.ndarray
        Normalized image data

    """
    im_min = image.min()
    im_max = image.max()
    return scale * ((image - im_min) / (im_max - im_min))


def get_weight_factor(size):
    """
    Calculates the weight factor for TV-denoising using scikit-image. Values have been determined empirically.

    Notes
    -----
    Formulas of fitting empirical data:

    - mean values: 3E-08x^2 + 8E-06x + 0,0037

    - more denoising: 3E-08x^2 + 6E-06x + 0,0127

    - linear fit was: 6E-05x - 0,0215

    Parameters
    ----------
    size: int
        Relevant size of image data
    Returns
    -------
    float
        TV-denoising weight factor
    """
    weight = 3e-8 * (size ** 2) + 8e-6 * size + 0.0037
    return weight


# TODO: inplace possible?
def denoise_image(image, denoise_weight=None):
    """
    Apply TV-denoising on input image data. Denoising weight factor is determined automatically, except explicitly
    specified

    Parameters
    ----------
    image : numpy.ndarray
        Input raw image data
    denoise_weight : int or float, optional
        Weight factor for scikit image TV-denoising

    Returns
    -------
    numpy.ndarray
        Denoised image data

    """
    # get weight factor
    small_edge_size = np.min(image.shape[:2])  # size of the smaller dim of an x-y image
    if denoise_weight:
        weight = denoise_weight
    else:
        weight = get_weight_factor(small_edge_size)

    # preallocate image array
    image_denoised = np.zeros_like(image, dtype=np.dtype('uint8'))

    # denoise per-plane for better results
    for z in range(image_denoised.shape[2]):
        # denoise_tv_chambolle always gives float64 in range [0,1]
        # rescale to unit8 [0,255] range
        image_denoised[:, :, z] = denoise_tv_chambolle(image[:, :, z], weight=weight) * 255

    #
    return image_denoised


def calc_weighted_median(df, data_col, weight_col):
    """
    Calculates the weighted median of specified DataFrame column with respect to another specified column-

    Parameters
    ----------
    df: pandas.DataFrame
        Input DataFrame that came from initial pore detection function
    data_col: str
        Name of column from which the median is calculated
    weight_col: str
        Name of column that is used as weight

    Returns
    -------
    pandas.Series
        Pandas Series with calculated weighted median

    """
    tmp_df = df.sort_values(data_col)
    cumsum = tmp_df[weight_col].values.cumsum()
    cutoff = tmp_df[weight_col].values.sum() / 2.0
    return tmp_df[data_col][cumsum >= cutoff].iloc[0]


def get_voxel_volume(sampling):
    return sampling['x'] * sampling['y'] * sampling['z']


# construct list of intervals as numpy array
def get_interval_list(part_size, number_parts):
    return [(np.array([0, 1]) + n) * part_size for n in range(number_parts)]


def get_intervals(data_shape, sampling, part_size_micron):
    # only depends on the part_size, the part numbers are calculated based on micron sizes
    number_parts_x, number_parts_y, number_parts_z = get_possible_part_numbers(data_shape, sampling, part_size_micron)
    part_size_x_px, part_size_y_px, part_size_z_px = get_part_size_px(data_shape, number_parts_x, number_parts_y,
                                                                      number_parts_z)

    intervals_x_px = get_interval_list(part_size_x_px, number_parts_x)
    intervals_y_px = get_interval_list(part_size_y_px, number_parts_y)
    intervals_z_px = get_interval_list(part_size_z_px, number_parts_z)

    return intervals_x_px, intervals_y_px, intervals_z_px


def get_cube_size_in_micron(data_shape, sampling):
    sx = int(data_shape[0]) * float(sampling['x'])
    sy = int(data_shape[1]) * float(sampling['y'])
    sz = int(data_shape[2]) * float(sampling['z'])
    return sx, sy, sz


def get_possible_part_numbers(data_shape, sampling, part_size):
    sx, sy, sz = get_cube_size_in_micron(data_shape, sampling)
    nx = np.floor(sx / part_size).astype('int')
    ny = np.floor(sy / part_size).astype('int')
    nz = np.floor(sz / part_size).astype('int')
    return nx, ny, nz


def get_part_size_px(data_shape, nx, ny, nz):
    px = np.floor(int(data_shape[0]) / nx).astype('int')
    py = np.floor(int(data_shape[1]) / ny).astype('int')
    pz = np.floor(int(data_shape[2]) / nz).astype('int')
    return px, py, pz



# Copyright (C) 2022 Tony Fischer
#
# This file is part of Bio Matrix Topology (BiMaTo).
#
# Bio Matrix Topology (BiMaTo) is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Bio Matrix Topology (BiMaTo) is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Bio Matrix Topology (BiMaTo).  If not, see <http://www.gnu.org/licenses/>.

"""
docstring of module
"""


__author__ = "Tony Fischer (tku137)"
__copyright__ = "Copyright 2022, Tony Fischer (tku137)"
__license__ = "GPLv3"
__email__ = "tonyfischer@mailbox.org"
__status__ = "Development"
__version__ = "2022.1"
__credits__ = ["Tony Fischer (tku137)", "Alexander Hayn"]

import numpy as np
from skimage.morphology import ball, remove_small_objects, skeletonize_3d
from skimage.measure import label


def get_network_skeleton(binary, min_size=10):
    """
    Calculates the network skeletonization of an input network binary.

    Parameters
    ----------
    binary : numpy.ndarray
        Input binary segmentation of typed `bool`
    min_size : int
        Minimum size of objects in `px`

    Returns
    -------
    numpy.ndarray
        Skeleton output

    """
    sk = skeletonize_3d(binary).astype(np.bool)
    sk_refined = remove_small_objects(sk, min_size=min_size, connectivity=3)
    return sk_refined


def classify_network(skeleton):
    """
    Classifies each binary fibril pixel of input skeleton depending on number of neighbouring pixels.

    Class 0 is fluid phase. Class 2 is a fibril with class 1 being a fibril ending. Classes higher than 2 mean that this
    is a node, while denoting the number of connections outgoing.

    Warning: the center pixel is always 1, so the sum of each cut out fibril environment needs to be reduced by 1

    Parameters
    ----------
    skeleton : numpy.ndarray
        Input skeleton

    Returns
    -------
    numpy.ndarray
        Classification matrix

    """
    fibril_coords = np.argwhere(skeleton > 0)

    # filter border coordinates
    x_size, y_size, z_size = skeleton.shape
    if not (x_size == y_size == z_size):
        raise ValueError("image dimensions are not equal")

    fibril_coords = fibril_coords[np.all(fibril_coords > 0, axis=1), :]
    fibril_coords = fibril_coords[np.all(fibril_coords < (x_size-1), axis=1), :]

    fibril_class = np.zeros_like(skeleton, dtype=np.uint8)
    for fib in fibril_coords:
        x, y, z = fib
        fib_cut = skeleton[x - 1:x + 2, y - 1:y + 2, z - 1:z + 2]
        fibril_class[x, y, z] = np.sum(fib_cut) - 1

    return fibril_class


def get_nodes(fibril_class, se_radius=1):
    # node analysis
    node_coords = np.argwhere(fibril_class > 2)  # coordinates of 'nodes' aka fibril_class > 2

    node_stack = np.zeros_like(fibril_class)  # empty stack
    for row in node_coords:  # for each node coordinate, draw ball
        x, y, z = row.astype('int')
        node_indices = np.argwhere(ball(se_radius))
        node_indices[:, 0] += x - se_radius
        node_indices[:, 1] += y - se_radius
        node_indices[:, 2] += z - se_radius

        # filter edge coords
        node_indices = node_indices[node_indices[:, 0] < node_stack.shape[0]]
        node_indices = node_indices[node_indices[:, 1] < node_stack.shape[1]]
        node_indices = node_indices[node_indices[:, 2] < node_stack.shape[2]]

        # binarize node stack
        node_stack[node_indices[:, 0], node_indices[:, 1], node_indices[:, 2]] = 1

    return node_stack


def get_single_fibrils_labeled(skeleton, node_stack):
    """
    Returns a label matrix with single fibrils. Each integer number in the label matrix represents pixels of a single,
    individual fibril.

    Parameters
    ----------
    fibril_class : numpy.ndarray
        Matrix with classified skeleton

    Returns
    -------
    numpy.ndarray
        Labelled matrix with single fibrils

    """
    # skeleton * inverse of node stack yields single fibril pixels
    single_fibrils = label(skeleton * np.logical_not(node_stack))

    return single_fibrils


def get_edge_fibril_ids(fibril_labels):
    edge_fibril_ids = np.unique(np.concatenate((
        np.unique(fibril_labels[:2, :, :]),  # front
        np.unique(fibril_labels[-2:, :, :]),  # back

        np.unique(fibril_labels[:, :2, :]),  # left
        np.unique(fibril_labels[:, -2:, :]),  # right

        np.unique(fibril_labels[:, :, :2]),  # bottom
        np.unique(fibril_labels[:, :, -2:]),  # top
    )))
    return edge_fibril_ids



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bimato']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.23.2,<2.0.0',
 'pandas>=1.4.3,<2.0.0',
 'scikit-image>=0.19.3,<0.20.0',
 'scipy>=1.9.0,<2.0.0']

setup_kwargs = {
    'name': 'bimato',
    'version': '2022.1',
    'description': 'Bio Matrix Topology (BiMaTo) is a library containing all the biopolymer matrix topology analyses published by the Biological Physics Group, (BIP), Peter Debye Institute, University Leipzig, Germany.',
    'long_description': 'None',
    'author': 'Tony Fischer (tku137)',
    'author_email': 'tonyfischer@mailbox.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<3.11',
}


setup(**setup_kwargs)
